@@ -17,9 +17,11 @@ from numpy import (array, asarray, cos, exp, linspace, matrix, meshgrid,
                    ndarray, ones, outer, real, remainder, sin, size, sqrt,
                    zeros_like)
 from scipy.signal import fftconvolve
+import plotly.graph_objects as go
+from plotly.subplots import make_subplots
 
 from . import np, degrees, eps
-from .utils import nearest2
+from .utils import nearest2, obj_2_xyz, azel_2_xyz
 
 # print(matplotlib.__version__)
 
@@ -33,406 +35,15 @@ name_colors = list(colors)
 #               ('dashdotdotted', (0, (3, 5, 1, 5, 1, 5))),
 #               ('loosely dashdotdotted', (0, (3, 10, 1, 10, 1, 10))),
 #               ('densely dashdotdotted', (0, (3, 1, 1, 1, 1, 1)))]
-linestyles = ['dotted', 'dashed', 'dashdot', 'loosely dotted', 'loosely dashed', 'loosely dashdot']
+linestyles = [
+    'dotted', 'dashed', 'dashdot', 'loosely dotted', 'loosely dashed',
+    'loosely dashdot'
+]
 
-# def draw2D(image, x, y, color="hot"):  # YlGnBu  RdBu, jet, hot
-#     """Draws a 2D image.
-#
-#     Parameters:
-#         x (numpy.array): x
-#         y (numpy.array): y
-#         xlabel (str): string for x label
-#         ylabel (str): string for y label
-#         title (str): title
-#         color (str): color
-#     """
-#
-#     id_fig = plt.figure()
-#     IDax = id_fig.add_subplot(111)
-#
-#     IDimage = plt.imshow(image,
-#                          interpolation='bilinear',
-#                          aspect='auto',
-#                          origin='lower',
-#                          extent=[x.min(), x.max(),
-#                                  y.min(), y.max()])
-#
-#     plt.colorbar()
-#     IDimage.set_cmap(color)
-#     return id_fig, IDax, IDimage
-
-
-def draw_empty_sphere(ax, angle_view, axis_equal=True):
-    """Fucntion that plots an empty Poincare sphere.
-    """
-
-    elev, azim = angle_view
-    max_size = 1
-
-    u = np.linspace(0, 360 * degrees, 90)
-    v = np.linspace(0, 180 * degrees, 90)
-
-    x = 1 * np.outer(np.cos(u), np.sin(v))
-    y = 1 * np.outer(np.sin(u), np.sin(v))
-    z = 1 * np.outer(np.ones_like(u), np.cos(v))
-
-    ax.plot_surface(x,
-                    y,
-                    z,
-                    rstride=4,
-                    cstride=4,
-                    color='b',
-                    edgecolor="k",
-                    linewidth=.0,
-                    alpha=0.1)
-    ax.plot(np.sin(u),
-            np.cos(u),
-            0,
-            color='k',
-            linestyle='dashed',
-            linewidth=0.5)
-    ax.plot(np.sin(u),
-            np.zeros_like(u),
-            np.cos(u),
-            color='k',
-            linestyle='dashed',
-            linewidth=0.5)
-    ax.plot(np.zeros_like(u),
-            np.sin(u),
-            np.cos(u),
-            color='k',
-            linestyle='dashed',
-            linewidth=0.5)
-
-    ax.plot([-1, 1], [0, 0], [0, 0], 'k-.', lw=1, alpha=0.5)
-    ax.plot([0, 0], [-1, 1], [0, 0], 'k-.', lw=1, alpha=0.5)
-    ax.plot([0, 0], [0, 0], [-1, 1], 'k-.', lw=1, alpha=0.5)
-
-    ax.plot(xs=(1, ),
-            ys=(0, ),
-            zs=(0, ),
-            color='black',
-            marker='o',
-            markersize=4,
-            alpha=0.5)
-
-    ax.plot(xs=(-1, ),
-            ys=(0, ),
-            zs=(0, ),
-            color='black',
-            marker='o',
-            markersize=4,
-            alpha=0.5)
-    ax.plot(xs=(0, ),
-            ys=(1, ),
-            zs=(0, ),
-            color='black',
-            marker='o',
-            markersize=4,
-            alpha=0.5)
-    ax.plot(xs=(0, ),
-            ys=(-1, ),
-            zs=(0, ),
-            color='black',
-            marker='o',
-            markersize=4,
-            alpha=0.5)
-    ax.plot(xs=(0, ),
-            ys=(0, ),
-            zs=(1, ),
-            color='black',
-            marker='o',
-            markersize=4,
-            alpha=0.5)
-    ax.plot(xs=(0, ),
-            ys=(0, ),
-            zs=(-1, ),
-            color='black',
-            marker='o',
-            markersize=4,
-            alpha=0.5)
-    distance = 1.15
-    ax.text(distance, 0, 0, 'Q', fontsize=14)
-    ax.text(0, distance, 0, 'U', fontsize=14)
-    ax.text(0, 0, distance, 'V', fontsize=14)
-
-    ax.view_init(elev=elev / degrees, azim=azim / degrees)
-
-    ax.set_xlabel('$S_1$', fontsize=14, labelpad=-10)
-    ax.set_ylabel('$S_2$', fontsize=14, labelpad=-10)
-    ax.set_zlabel('$S_3$', fontsize=14, labelpad=-10)
-    ax.grid(False)
-
-    ax.set_xticklabels([])
-    ax.set_yticklabels([])
-    ax.set_zticklabels([])
-
-    ax.set_xlim(-max_size, max_size)
-    ax.set_ylim(-max_size, max_size)
-    ax.set_zlim(-max_size, max_size)
-
-    plt.tight_layout()
-    # set_aspect_equal_3d(ax)
-    if axis_equal:
-        try:
-            ax.set_aspect('equal')
-        except:
-            pass
-
-
-def draw_poincare(S,
-                  normalize=True,
-                  remove_depol=False,
-                  kind='scatter',
-                  color_scatter='r',
-                  color_line='r',
-                  angle_view=[0.5, -1],
-                  figsize=(6, 6),
-                  filename='',
-                  subplots=None,
-                  axis_equal=True):
-    """Function to draw the poincare sphere.
-    It admits Stokes or a list with Stokes, or None
-
-    Parameters:
-        stokes_points (Stokes, list, None): list of Stokes points.
-        angle_view (float, float): Elevation and azimuth for viewing.
-        is_normalized (bool): normalize intensity to 1.
-        kind (str): 'line', 'scatter' or 'both'.
-        color (str): color of line or scatter.
-        label (str): labels for drawing.
-        filename (str): name of filename to save the figure.
-        axis_equal (bool): If True, the axis_equal method is used. Default: True.
-    """
-    # Calculate the number of subplots
-    if subplots is None:
-        Nx, Ny, Nsubplots, Ncurves = (1, 1, 1, S.size)
-    elif type(subplots) is tuple:
-        Nsubplots = np.prod(np.array(subplots[0:2]))
-        if S.size % Nsubplots != 0:
-            raise ValueError(
-                'Shape {} is not valid for the object {} of {} elements'.
-                format(subplots, S.name, S.size))
-        Nx, Ny = subplots[0:2]
-        Ncurves = int(S.size / Nsubplots)
-    elif subplots.upper() == 'INDIVIDUAL':
-        Ny = int(np.floor(np.sqrt(S.size)))
-        Nx = int(np.ceil(S.size / Ny))
-        Nsubplots = S.size
-        Ncurves = 1
-    elif subplots.upper() == 'AS_SHAPE':
-        if S.ndim < 2:
-            Nx, Ny = (1, S.size)
-            Nsubplots, Ncurves = (S.size, 1)
-        else:
-            Nx, Ny = S.shape[0:2]
-            Nsubplots = Nx * Ny
-        Ncurves = int(S.size / Nsubplots)
-    else:
-        raise ValueError('{} is not a valid subplots option.')
-
-    # Flatten the object
-    Sf = S.copy()
-    Sf.shape = np.array([Sf.size])
-
-    # Calculate color min and max values
-    if color_line.upper() == 'INTENSITY' or color_scatter.upper(
-    ) in 'INTENSITY':
-        I = Sf.parameters.intensity(out_number=False)
-        I_min, I_max = (I.min(), I.max())
-    else:
-        I_min, I_max = (None, None)
-    if color_line.upper() == 'DEGREE' or color_scatter.upper() in 'DEGREE':
-        pol = Sf.parameters.degree_polarization(out_number=False)
-        D_min, D_max = (pol.min(), pol.max())
-    else:
-        D_min, D_max = (None, None)
-
-    # Create the figure
-    fig = plt.figure(figsize=figsize)
-    ax = []
-
-    # Loop to generate the poincare spheres
-    for indS in range(Nsubplots):
-        # Divide in subplots
-        axis = fig.add_subplot(Nx,
-                               Ny,
-                               indS + 1,
-                               projection='3d',
-                               adjustable='box')
-        ax += [axis]
-        # Create the Poincare sphere
-        draw_empty_sphere(axis, angle_view=angle_view, axis_equal=axis_equal)
-
-        # Add points from Stokes
-        im_scatter = draw_stokes_points(axis,
-                                        Sf[indS * Ncurves:(indS + 1) *
-                                           Ncurves],
-                                        I_min=I_min,
-                                        I_max=I_max,
-                                        D_min=D_min,
-                                        D_max=D_max,
-                                        normalize=normalize,
-                                        remove_depol=remove_depol,
-                                        kind=kind,
-                                        color_scatter=color_scatter,
-                                        color_line=color_line)
-        # Add titles
-        if Nsubplots > 1:
-            if subplots in ('individual', 'Individual', 'INDIVIDUAL'):
-                string = str(indS)
-            else:
-                string = str(list(np.unravel_index(indS, (Nx, Ny))))
-            plt.title(string, fontsize=18)
-        else:
-            plt.title(S.name, fontsize=26)
-
-    # Add supertitle if required
-    if Nsubplots > 1:
-        fig.suptitle(S.name, fontsize=26)
-
-    # Add colormap if required
-    if im_scatter is not None:
-        cbar = fig.colorbar(im_scatter, ax=ax)
-        if color_scatter.upper() == 'INTENSITY':
-            cbar.set_label(label='Intensity', fontsize=14)
-        else:
-            cbar.set_label(label='Pol. degree', fontsize=14)
-
-    return ax, fig
-
-
-def draw_stokes_points(ax,
-                       S,
-                       I_min=None,
-                       I_max=None,
-                       D_min=None,
-                       D_max=None,
-                       normalize=True,
-                       remove_depol=False,
-                       kind='scatter',
-                       color_scatter='Intensity',
-                       color_line='r'):
-    """Function to draw Stokes vectors on the poincare sphere.
-
-    Parameters:
-        S (Stokes): Stokes object.
-        normalize (bool): If True, normalize the Stokes vectors to have intensity 1. Default: True.
-        I_min (float): Minimum intensity among all subplots.
-        I_max (float): Amximum intensity among all subplots.
-        D_min (float): Minimum polarization degree among all subplots.
-        D_max (float): Amximum polarization degree among all subplots.
-        remove_depol (bool): If True, plot the polarized part of the Stokes vector. Default: False.
-        kind (str): Plot type: LINE, SCATTER or BOTH. Default: SCATTER.
-        color_scatter (str): There are three options. INTENSITY sets the color of the points intensity dependent. DEGREE sets the color to match the polarization degree. Another posibility is to use valid color strings such as 'r'. Default: INTENSITY.
-        color_line (str): Only valid color strings such as 'r'. Default: 'r'.
-
-    Returns:
-
-    """
-    # Avoid empty objects
-    if S is None or S.size == 0:
-        return None
-
-    # if there is just one item in the Stokes object, change to scatter
-    if S.size == 1:
-        kind = 'scatter'
-
-    # Remove deopolarization if required
-    if remove_depol:
-        Sp, _ = S.parameters.polarized_unpolarized()
-    else:
-        Sp = S
-
-    # Normalize to intensity=1 if required
-    I = Sp.parameters.intensity(shape=False, out_number=False)
-    if normalize:
-        Sn = Sp.normalize(keep=True)
-    else:
-        Sn = Sp
-
-    # Calculate all the required parameters
-    _, S1, S2, S3 = Sn.parameters.components(shape=False, out_number=False)
-    im_scatter = None
-    if color_scatter.upper() == 'DEGREE':
-        depol = Sn.parameters.degree_depolarization(shape=False,
-                                                    out_number=False)
-
-    # Make the plots
-    if kind.upper() in ('SCATTER', 'BOTH'):
-        if color_scatter.upper() == 'INTENSITY':
-            im_scatter = ax.scatter(S1,
-                                    S2,
-                                    S3,
-                                    c=I,
-                                    s=60,
-                                    vmin=I_min,
-                                    vmax=I_max)
-        elif color_scatter.upper() == 'DEGREE':
-            im_scatter = ax.scatter(S1,
-                                    S2,
-                                    S3,
-                                    c=depol,
-                                    s=60,
-                                    vmin=D_min,
-                                    vmax=D_max)
-        else:
-            ax.scatter(S1, S2, S3, c=color_scatter, s=60)
-
-    if kind.upper() in ('LINE', 'BOTH') and Sn.size > 1:
-        ax.plot(S1, S2, S3, c=color_line, lw=2)
-
-    return im_scatter
-    #
-    #
-    #
-    # # Plot the points
-    # if stokes_points is not None:
-    #     if isinstance(stokes_points, list):
-    #         if isinstance(stokes_points[0], matrix):
-    #             stokes_points = asarray(stokes_points)
-    #             s0 = stokes_points[:, 0].squeeze()
-    #             s1 = stokes_points[:, 1].squeeze()
-    #             s2 = stokes_points[:, 2].squeeze()
-    #             s3 = stokes_points[:, 3].squeeze()
-    #         else:
-    #             points = []
-    #
-    #             for i, point in enumerate(stokes_points):
-    #                 points.append(point.M)
-    #
-    #             stokes_points = asarray(points)
-    #             s0 = array(stokes_points[:, 0]).squeeze()
-    #             s1 = array(stokes_points[:, 1]).squeeze()
-    #             s2 = array(stokes_points[:, 2]).squeeze()
-    #             s3 = array(stokes_points[:, 3]).squeeze()
-    #     elif isinstance(stokes_points, ndarray):
-    #         s0 = array(stokes_points[:, 0]).squeeze()
-    #         s1 = array(stokes_points[:, 1]).squeeze()
-    #         s2 = array(stokes_points[:, 2]).squeeze()
-    #         s3 = array(stokes_points[:, 3]).squeeze()
-    #     else:
-    #         s0 = array(stokes_points.M[0]).squeeze()
-    #         s1 = array(stokes_points.M[1]).squeeze()
-    #         s2 = array(stokes_points.M[2]).squeeze()
-    #         s3 = array(stokes_points.M[3]).squeeze()
-    #
-    #     if is_normalized is True:
-    #         s1 = s1 / s0
-    #         s2 = s2 / s0
-    #         s3 = s3 / s0
-    #         max_size = 1
-    #     else:
-    #         max_size = s0.max()
-    #
-    # if kind == 'line':
-    #     ax.plot(s1, s2, s3, c=color, lw=2, label=label)
-    # elif kind == 'scatter':
-    #     ax.scatter(s1, s2, s3, c=color, s=60, label=label)
-    #
-    # plt.tight_layout()
-    # if filename not in (None, [], ''):
-    #     plt.savefig(filename)
+# Use only when creating docs
+# import plotly.io as pio
+# pio.renderers.default = "jupyterlab"
+# print(pio.renderers)
 
 
 def draw_ellipse(E,
@@ -446,23 +57,23 @@ def draw_ellipse(E,
                  depol_prob=False,
                  subplots=None,
                  N_prob=256,
-                 contour_levels=(0.9, ),
+                 contour_levels=0.9,
                  cmap='hot'):
     """Draws polarization ellipse of Jones vector.
 
     Parameters:
         E (Jones_vector or Stokes): Light object.
         N_angles (int): Number of angles to plot the ellipses. Default: 91.
-        limit (float): limit for drawing. If empty, it is obtained from ampltiudes.
         filename (str): name of filename to save the figure.
         figsize (tuple): A tuple of length 2 containing the figure size. Default: (8,8).
+        limit (float): limit for drawing. If empty, it is obtained from amplitudes.
         draw_arrow (bool): If True, draws an arrow containing the turning sense of the polarization. Does not work with linear polarization vectors. Default: True.
         depol_central (bool): If True, draws a central circle containing the unpolarized field amplitude. Default: False.
-        depol_contour (bool): If True, draws a line enveloping the polarization ellipse in ordeer to plot the depolarization. Default: False.
-        depol_dist (bool): If True, plots the probability distribution of the electric field. Default: False.
+        depol_contour (bool): If True, draws a line enveloping the polarization ellipse in order to plot the depolarization. Default: False.
+        depol_prob (bool): If True, plots the probability distribution of the electric field. Default: False.
         subplots (string, tuple or None): If AS_SHAPE, divides the figure in several subplots as the shape of the py_pol object. If INDIVIDUAL, each vector is represented in its own subaxis, trying to use a square grid. If tuple, divides the figure in that same number of subplots. If None, all ellipses are plot in the same axes. Default: None.
         N_prob (int): Number of points in each dimension for probability distributions. Default: 256.
-        prob (flota, np.ndarray, tuple or list): Contains the contour levels (normalized to 1). Default: 0.9.
+        contour_levels (float, np.ndarray, tuple or list): Contains the contour levels (normalized to 1). Default: 0.9.
         cmap (str or color object): Default colormap for probability distributions. Default: hot.
 
     Returns:
@@ -490,9 +101,8 @@ def draw_ellipse(E,
     if E._type == 'Jones_vector':
         is_linear = E.checks.is_linear(shape=False, out_number=False)
     else:
-        is_linear = E.checks.is_linear(shape=False,
-                                       out_number=False,
-                                       use_nan=False)
+        is_linear = E.checks.is_linear(
+            shape=False, out_number=False, use_nan=False)
     # Create the electric field distributions
     Ex = E0X * np.cos(Angles + Phase)
     Ey = E0Y * np.cos(Angles + Phase + Delay)
@@ -610,31 +220,30 @@ def draw_ellipse(E,
                 string = str(list(np.unravel_index(ind, E.shape)))
         # Plot the probability distribution
         if depol_prob and is_depol[ind]:
-            IDimage = axis.imshow(prob[ind, :, :],
-                                  interpolation='bilinear',
-                                  aspect='equal',
-                                  origin='lower',
-                                  extent=[-limit, limit, -limit, limit])
+            IDimage = axis.imshow(
+                prob[ind, :, :],
+                interpolation='bilinear',
+                aspect='equal',
+                origin='lower',
+                extent=[-limit, limit, -limit, limit])
             # axis = axis[0]
         # Plot the curve
         if is_pol[ind]:
             axis.plot(Ex[ind, :], Ey[ind, :], lw=2, label=string, color=color)
             if draw_arrow and ~is_linear[ind]:
-                axis.arrow(Ex[ind, 0],
-                           Ey[ind, 0],
-                           Ex[ind, 4] - Ex[ind, 0],
-                           Ey[ind, 4] - Ey[ind, 0],
-                           width=0,
-                           head_width=0.075 * limit,
-                           linewidth=0,
-                           color=color,
-                           length_includes_head=True)
+                axis.arrow(
+                    Ex[ind, 0],
+                    Ey[ind, 0],
+                    Ex[ind, 4] - Ex[ind, 0],
+                    Ey[ind, 4] - Ey[ind, 0],
+                    width=0,
+                    head_width=0.075 * limit,
+                    linewidth=0,
+                    color=color,
+                    length_includes_head=True)
         elif E._type == 'Stokes' and depol_central and ~is_depol[ind]:
-            axis.plot(np.zeros(2),
-                      np.zeros(2),
-                      lw=2,
-                      label=string,
-                      color=color)
+            axis.plot(
+                np.zeros(2), np.zeros(2), lw=2, label=string, color=color)
         elif depol_central or depol_prob:
             print('Field {} is empty.'.format(string))
         else:
@@ -642,8 +251,7 @@ def draw_ellipse(E,
         # Add the depolarization for Stokes vectors
         if E._type == 'Stokes' and depol_central and is_depol[ind]:
             axis.plot(Exu[ind, :], Eyu[ind, :], lw=1.5, color=color, ls='--')
-        if E._type == 'Stokes' and depol_contour and is_pol[ind] and is_depol[
-                ind]:
+        if E._type == 'Stokes' and depol_contour and is_pol[ind] and is_depol[ind]:
             CS = axis.contour(
                 x,
                 x,
@@ -652,7 +260,7 @@ def draw_ellipse(E,
                 colors=(color),
                 linewidths=1.5,
                 linestyles=line_styles)
-                # linestyles=('dashdot'))
+            # linestyles=('dashdot'))
         # Additions to figure
         if indC == Ncurves - 1:
             plt.axis('equal')
@@ -797,3 +405,249 @@ def draw_ellipse(E,
 #     ax.set_xlim3d([xmean - factor * plot_radius, xmean + factor * plot_radius])
 #     ax.set_ylim3d([ymean - factor * plot_radius, ymean + factor * plot_radius])
 #     ax.set_zlim3d([zmean - 1 * plot_radius, zmean + 1 * plot_radius])
+
+
+def draw_poincare(S, fig=None, figsize=(6,6), draw_axes=True, draw_guides=True, kind="scatter", depol=False, param=None, subplots=False, in_degrees=False, log=False, hover="components", colormap="Blackbody", show_fig=False):
+    """Function to draw the Jones or Stokes vectors in the Poincare sphere using plotly.
+
+    Args:
+        S (Jones_vector or Stokes): Object to be represented.
+        fig (plotly.graph_objects.Figure): Figure to plot the data. Default: None
+        figsize (2-element iterable): Figure size. Default: (6,6)
+        draw_axes (bool): If True, it draws the three axes of the Poincare space. Default: True
+        draw_guides (bool): If True, it draws the circles of S1, S2 and S3 = 0. Default: True
+        kind (str): Choose between 'scatter', 'line', 'scatterline' or 'surf' to represent the data. If surf, the object must be 2D, 3D or 4D. Else, it must be 0D, 1D, 2D or 3D. Default: 'scatter'.
+        depol (bool): If True, the depolarization is taking into account for scatter and line plots shrinking the radius below 1. Default: False.
+        param (str, np.ndarray or None): If str, parameter to use as information for the color. Must be a method of Parameters class which returns a single variable. If np.ndarray, must have the same shape as S. Default: None.
+        subplots (bool): If True, it tries to use the first two dimensions as rows and columns. If method == 'surf', this is mandatory. Default: False.
+        in_degrees (bool): If True, transforms the parameters to degrees. Default: False.
+        log (bool): If True, it calculates it in logarithmic scale. Default: False.
+        hover (str): Choose between 'components' (S1, S2, S3) or 'angles' (azimuth, ellipticity). Default: 'components'.
+        colormap (str): Colormap of the plots. Default: 'Blackbody'.
+        show_fig (bool): If True, the figure is inmediately plotted. Default: False.
+
+    Returns:
+        fig (go.Figure): Plotly figure.
+    """
+    # Unrelated problems
+    if hover == "angles":
+        if kind == "surf":
+            hover = "components"
+            print("WARNING: Plotly has an issue in the hover of surfaces. Hover mode changed to components.")
+        else:
+            print("WARNING: Plotly has an issue in the hover of surfaces. Values are not correct.")
+    # Allow several figures. Indices: [row, column, trace, dots inside trace]
+    S = S.copy()
+    Ndim = S.ndim
+    if kind == "surf":
+        # Surf requires pseudo-1D, 2D, and higher dimensions to be separated in subplots
+        if Ndim > 4 or Ndim < 1:
+            raise ValueError("Object has a wrong number of dimensions ({})".format(Ndim))
+        elif Ndim == 1:
+            if param is None:
+                raise ValueError("Object has a too few dimensions ({})".format(Ndim))
+            else:
+                # TODO: Interpolate
+                Nrows = 1
+                Ncolumns = 1
+                S.shape = [1, 1, 1] + [S.size]
+        elif Ndim == 2:
+            Nrows = 1
+            Ncolumns = 1
+            S.shape = [1, 1, 1] + S.shape
+        elif Ndim == 3:
+            Nrows = 1
+            Ncolumns = S.shape[0]
+            S.shape = [S.shape[0]] + [1, 1] + S.shape[1:]
+        elif Ndim == 4:
+            Nrows = S.shape[0]
+            Ncolumns = S.shape[1]
+            S.shape = S.shape[0:2] + [1] + S.shape[2:]
+        Ones = np.ones(S.shape[3:])
+    elif subplots:
+        # When subplots is True, elements of 2D and 3D are separated in differentsubplots. Higher dimensions are taken as part of the 3rd dimension
+        if Ndim < 2:
+            Nrows = 1
+            Ncolumns = 1
+            S.shape = [1, 1, 1] + [S.size]
+        elif Ndim == 2:
+            Nrows = 1
+            Ncolumns = S.shape[0]
+            S.shape = [1] + [S.shape[0]] + [1] + [S.shape[1]]
+        elif Ndim == 3:
+            Nrows = S.shape[0]
+            Ncolumns = S.shape[1]
+            S.shape = S.shape[0:2] + [1] + [S.shape[2]]
+        else:
+            Nrows = S.shape[0]
+            Ncolumns = S.shape[1]
+            S.shape = S.shape[0:3] + [np.prod(S.shape[3:])]
+        Ones = np.ones(S.shape[3])
+    else:
+        # When subplots is false, all traces are plotted in the same sphere. Dimensions higher than 2D are included in the 2nd dimension.
+        Nrows = 1
+        Ncolumns = 1
+        if Ndim <= 1:
+            S.shape = [1, 1, 1, S.size]
+        else:
+            S.shape = [1, 1, S.shape[0]] + [np.prod(S.shape[1:])]
+        Ones = np.ones(S.shape[3])
+    # Reshape param
+    if param is not None and not isinstance(param, str):
+        param = np.reshape(param, S.shape)
+
+
+    # Figure and figure options
+    add_auxiliar = False
+    if fig is None:
+        specs = [[{'type': 'surface'}] * Ncolumns] * Nrows
+        fig = make_subplots(rows=Nrows, cols=Ncolumns, specs=specs)
+        add_auxiliar = True
+    lighting = dict(ambient=0.9, diffuse=0., roughness = 0.5, specular=0.05, fresnel=0.2)
+    if hover == 'components':
+        hovertemplate = "S1: %{x:.3f}<br>S2: %{y:.3f}<br>S3: %{z:.3f}"
+    else:
+        hovertemplate = "Az: %{customdata[0]:.1f}<br>El: %{customdata[1]:.1f}"
+
+    # Axes
+    if draw_axes and add_auxiliar:
+        line_p = dict(width=6, color="red")
+        line_m = dict(width=6, color="red", dash="dash")
+        marker = dict(size=1)
+        axis_px = go.Scatter3d(x=[0,1.2], y=[0,0], z=[0,0], line=line_p, marker=marker, hoverinfo="skip", name="$S_1$")
+        axis_mx = go.Scatter3d(x=[0,-1.2], y=[0,0], z=[0,0], line=line_m, marker=marker, hoverinfo="skip", name="$-S_1$")
+        line_p["color"], line_m["color"] = ("green", "green")
+        axis_py = go.Scatter3d(x=[0,0], y=[0,1.2], z=[0,0], line=line_p, marker=marker, hoverinfo="skip", name="$S_2$")
+        axis_my = go.Scatter3d(x=[0,0], y=[0,-1.2], z=[0,0], line=line_m, marker=marker, hoverinfo="skip", name="$-S_2$")
+        line_p["color"], line_m["color"] = ("blue", "blue")
+        axis_pz = go.Scatter3d(x=[0,0], y=[0,0], z=[0,1.2], line=line_p, marker=marker, hoverinfo="skip", name="$S_3$")
+        axis_mz = go.Scatter3d(x=[0,0], y=[0,0], z=[0,-1.2], line=line_m, marker=marker, hoverinfo="skip", name="$-S_3$")
+        size=15
+        annotations = [
+            dict(showarrow=False, x=1.2, y=0, z=0, text="$S_1$", font=dict(color="red", size=size), xshift=15, yshift=15),
+            dict(showarrow=False, x=-1.2, y=0, z=0, text="$-S_1$", font=dict(color="red", size=size), xshift=-15, yshift=-15),
+            dict(showarrow=False, x=0, y=1.2, z=0, text="$S_2$", font=dict(color="green", size=size), xshift=15, yshift=15),
+            dict(showarrow=False, x=0, y=-1.2, z=0, text="$-S_2$", font=dict(color="green", size=size), xshift=-15, yshift=-15),
+            dict(showarrow=False, x=0, y=0, z=1.2, text="$S_3$", font=dict(color="blue", size=size), xshift=15, yshift=15),
+            dict(showarrow=False, x=0, y=0, z=-1.2, text="$-S_3$", font=dict(color="blue", size=size), xshift=-15, yshift=-15),
+        ]
+    else:
+        annotations = []
+
+    # Curves
+    if draw_guides and add_auxiliar:
+        angle = np.linspace(0, 360*degrees, 361)
+        line = dict(width=2, color="darkslategrey", dash="dashdot")
+        x = np.sin(angle)
+        y = np.cos(angle)
+        z = np.zeros_like(angle)
+        circle_z = go.Scatter3d(x=x, y=y, z=z, line=line, mode="lines", name="S3=0")
+        x = np.sin(angle)
+        z = np.cos(angle)
+        y = np.zeros_like(angle)
+        circle_y = go.Scatter3d(x=x, y=y, z=z, line=line, mode="lines", name="S2=0")
+        y = np.sin(angle)
+        z = np.cos(angle)
+        x = np.zeros_like(angle)
+        circle_x = go.Scatter3d(x=x, y=y, z=z, line=line, mode="lines", name="S1=0")
+
+    # Poincare sphere
+    if kind != "surf" and add_auxiliar:
+        # az, el = np.mgrid[0:180*degrees:100j, -45*degrees:45*degrees:100j]
+        el, az = np.mgrid[-45*degrees:45*degrees:100j, 0:180*degrees:100j]
+        x, y, z = azel_2_xyz(az, el)
+        customdata = np.dstack((az/degrees, el/degrees)) if hover == "angles" else []
+        level = 0.2
+        Psphere = go.Surface(x=x, y=y, z=z, surfacecolor=np.ones_like(x)*level, cmin=0, cmax=1, opacity=0.5,
+            colorscale="Greys", showscale=False, lighting=lighting, customdata=customdata,
+            hovertemplate=hovertemplate, name="Sphere", showlegend=False)
+
+    # Loop in rows
+    for indR, Srow in enumerate(S):
+        for indC, Scol in enumerate(Srow):
+
+            # Axes
+            if draw_axes and add_auxiliar:
+                fig.add_traces([axis_px, axis_mx, axis_py, axis_my, axis_pz, axis_mz], rows=indR+1, cols=indC+1)
+
+            # Curves
+            if draw_guides and add_auxiliar:
+                fig.add_traces([circle_z, circle_y, circle_x], rows=indR+1, cols=indC+1)
+
+            # Loop in traces
+            for indT, Strace in enumerate(Scol):
+
+                # Parameter
+                if isinstance(param, str):
+                    colorbar = dict(title=param)
+                    Scolor = eval("Strace.parameters." + param + "(out_number=False)")
+                elif param is not None:
+                    colorbar = dict(title=S.name)
+                    Scolor = param[indR, indC, indT, ...]
+                else:
+                    colorbar = {}
+                    Scolor = Ones * (indT + 1) / Scol.shape[0]
+                if param is not None:
+                    if log:
+                        cond = Scolor <= 0
+                        if np.any(cond):
+                            Scolor[cond] = np.inf
+                            Scolor[cond] = np.min(Scolor) / 10
+                        Scolor = np.log10(Scolor)
+                        colorbar["title"] = "log(" + colorbar["title"] + ")"
+                    elif in_degrees:
+                        Scolor = Scolor / degrees
+                        colorbar["title"] += " (deg)"
+
+                # Plot data
+                if kind != "surf":
+                    # Poincare sphere
+                    if add_auxiliar:
+                        fig.add_trace(Psphere, row=indR+1, col=indC+1)
+
+                    # Data
+                    if kind in ("scatter", "scatterline"):
+                        x, y, z, az, el, _ = obj_2_xyz(Strace, in_degrees=True, depol=depol)
+                        customdata = np.squeeze(np.dstack((az, el))) if hover == "angles" else []
+                        if kind == "scatter":
+                            marker = dict(size=10, color=Scolor, colorbar=colorbar, colorscale=colormap)
+                        else:
+                            marker = dict(size=10, color=Scolor, colorscale=colormap)
+                        Fdata = go.Scatter3d(x=x, y=y, z=z, marker=marker, name=S.name,
+                                        mode="markers", customdata=customdata, hovertemplate=hovertemplate)
+                        # TODO: Should be possible to link colorbars. Not right now due to a bug in Plotly.
+                        # marker = dict(size=10, color=Scolor, colorscale=colormap)
+                        # Fdata = go.Scatter3d(x=x, y=y, z=z, marker=marker, name=S.name,
+                        #                 mode="markers", customdata=customdata, hovertemplate=hovertemplate, coloraxis="coloraxis")
+                        fig.add_trace(Fdata, row=indR+1, col=indC+1)
+                    if kind in ("line", "scatterline"):
+                        x, y, z, az, el, Scolor = obj_2_xyz(Strace, DAinterp=1*degrees, in_degrees=True, param=Scolor, depol=depol)
+                        customdata = np.squeeze(np.dstack((az, el))) if hover == "angles" else []
+                        line = dict(width=8, color=Scolor, colorbar=colorbar, colorscale=colormap)
+                        Fdata = go.Scatter3d(x=x, y=y, z=z, line=line, name=S.name+" int.",
+                                        mode="lines", customdata=customdata, hovertemplate=hovertemplate)
+                        fig.add_trace(Fdata, row=indR+1, col=indC+1)
+
+                else:
+                    x, y, z, az, el, Scolor = obj_2_xyz(Strace, in_degrees=True, param=Scolor, interp_to_surf=True)
+                    # customdata = [az[0,:], el[:, 0]] if hover == "angles" else None
+                    customdata = np.squeeze(np.stack((az, el), axis=-1)) if hover == "angles" else None
+                    # print(customdata[0], customdata[1])
+                    Psphere = go.Surface(x=x, y=y, z=z, surfacecolor=Scolor, opacity=1,
+                        colorscale=colormap, lighting=lighting, customdata=customdata,
+                        hovertemplate=hovertemplate, name=S.name, colorbar=colorbar)
+                    fig.add_trace(Psphere, row=indR+1, col=indC+1)
+
+
+    # Plot figure
+    axis = dict(showbackground=False, showgrid=False, zeroline=False, visible=False)
+    camera = dict(up=dict(x=0, y=0, z=1), center=dict(x=0, y=0, z=0), eye=dict(x=0.75, y=0.75, z=0.75))
+    fig.update_layout(margin=dict(l=20, r=20, t=20, b=20), width=int(figsize[0]*100),
+                    height=int(figsize[1]*100), showlegend=False)
+    fig.update_scenes(annotations=annotations, xaxis=axis, yaxis=axis, zaxis=axis, camera=camera)
+    # TODO: Annotations do not render. Why???
+
+    if show_fig:
+        fig.show()
+
+    return fig
