@@ -14,6 +14,7 @@ import time
 import matplotlib.pyplot as plt
 import numpy as np
 import scipy as sp
+from scipy.interpolate import interp2d, griddata
 from numpy import array, cos, matrix, pi, sin, sqrt, tan
 from scipy.optimize import leastsq
 
@@ -555,9 +556,9 @@ def take_shape(objs):
     """Calculates the shape with higher dimensionality."""
     if isinstance(objs, (list, tuple)):
         l = 0
-        shape = None
+        shape = [0]
         for obj in objs:
-            # Fifferenciate between lists of objects and shapes
+            # Differenciate between lists of objects and shapes
             if isinstance(obj, (list, tuple, np.ndarray, NoneType)):
                 s = obj
             else:
@@ -566,6 +567,8 @@ def take_shape(objs):
                 if len(s) > l:
                     l = len(s)
                     shape = s
+                elif len(s) == l and np.prod(s) > np.prod(shape):
+                    shape = s
     else:
         shape = objs.shape
     return shape
@@ -797,16 +800,27 @@ def PrintMatrices(list, Nspaces=3):
     Returns:
         str (string): String to be printed.
     """
+    # Safety
+    if list is None or list == []:
+        return "This object is empty"
+
+    # Get size and shape
     N = len(list)
-    s = list[0].shape
+    shape = list[0].shape
     str = ''
-    for r in range(s[0]):
+    # Iterate in rows
+    for row in range(shape[0]):
         for M in list:
+            # Check if the printed numbers are complex
+            cond_complex = M.dtype in (np.complex64, np.complex128) and np.sum(M.imag**2)/M.size > tol_default
             # Start of matrix
             str = str + '['
-            # Row of matrix
-            for c in range(s[1]):
-                str = str + '{:+1.3f} '.format(M[r, c])
+            # Print full row of matrix
+            for col in range(shape[1]):
+                if cond_complex:
+                    str = str + '{:+1.3f} + {:+1.3f}j '.format(M[row, col].real, M[row, col].imag)
+                else:
+                    str = str + '{:+1.3f} '.format(M[row, col].real)
             # End of matrix
             str = str + '\b]'
             str = str + Nspaces * ' '
@@ -1174,8 +1188,9 @@ def put_in_limits(x, typev, out_number=True):
 
     elif typev in ("azimuth", "Az", 'azimuth', 'Azimuth'):
         cond = np.logical_or(x < limAz[0], x >= limAz[1])
-        aux = cos(x[cond])
-        x[cond] = np.arccos(-aux)
+        # aux = cos(x[cond])
+        # x[cond] = np.arccos(aux)
+        x[cond] = x[cond] % np.pi
 
     elif typev in ("ellipticity", "El", 'ellipticity', 'Ellipticity'):
         cond = np.logical_or(x < limEl[0], x >= limEl[1])
@@ -1185,9 +1200,9 @@ def put_in_limits(x, typev, out_number=True):
         aux[cond2] = 1 / aux[cond2]
         x[cond] = np.arctan(aux)
     elif typev in ('ret', 'Ret', 'RET', 'Retardance', 'retardance'):
-        cond = np.logical_or(x < limRet[0], x >= limRet[1])
+        cond = np.logical_or(x < limRet[0], x > limRet[1])
         aux = cos(x[cond])
-        x[cond] = np.arccos(-aux)
+        x[cond] = np.arccos(aux)
     # Recover the nan values
     if np.any(condNan):
         x[condNan] = np.nan
@@ -1842,4 +1857,93 @@ def fit_sine(t, data, has_draw=True):
     return est_amp, est_freq, est_phase, est_mean
 
 
+def obj_2_xyz(S, Ninterp=1, DAinterp=None, in_degrees=False, param=None, interp_to_surf=False, depol=False):
+    """Function to calculate the x, y and z coordinates of the Poincare sphere.
+
+    Args:
+        S (Jones_vector or Stokes): 0D or 1D object to be represented.
+        Ninterp (number): Multiplication factor to interpolate angles if DAinterp is not None. Default: 1.
+        DAinterp (number): Angle step for interpolation. Default: None.
+        return_azel (bool): If True, returns interpolated azimuth and ellipticity. Default: False.
+        in_degrees (bool): If True, azimuth and ellipticity are returned in degrees. Default: False.
+        interp_to_surf (bool): If True and S is 1D, transform it to 2D. Default: True.
+        depol (bool): If True, the depolarization is taking into account for scatter and line plots shrinking the radius below 1. Default: False.
+
+    Returns:
+        x, y, z (np.ndarray): Calculated coordinates.
+        az, el (np.ndarray): Interpolated azimuth and ellipticity.
+        param (np.ndarray): Calculated and interpolated parameter.
+    """
+    # Calculate actual spherical coordinates
+    az, el = S.parameters.azimuth_ellipticity(out_number=False, use_nan=False)
+    if depol:
+        pol = S.parameters.degree_polarization(out_number=False, use_nan=False)
+        pol[np.isnan(pol)] = eps
+        pol[pol < eps] = eps
+        pol[pol > 1] = 1
+    else:
+        pol = 1
+
+    # Interpolate 1D data to plot in the 2D sphere
+    if interp_to_surf and S.ndim == 1:
+        # Calculate spherical coordinates
+        az_int = np.linspace(0, 180*degrees, int(np.ceil(179.999*degrees/(DAinterp or 1*degrees))+1))
+        el_int = np.linspace(-45*degrees, 45*degrees, int(np.ceil(90*degrees/(DAinterp or 1*degrees))+1))
+        Az_int, El_int = np.meshgrid(az_int, el_int)
+        # Calculate interpolation
+        param = griddata((az, el), param, (Az_int, El_int), method='cubic')
+        az, el = Az_int, El_int
+
+    # Interpolate scatter points to plot the line along the sphere
+    else:
+        # Interpolate in angles step
+        if DAinterp is not None:
+            Daz, Del = (np.diff(az), np.diff(el))
+            Dstep = np.maximum(Daz, Del)
+            Nstep = np.array(np.ceil(Dstep/DAinterp), dtype=int)
+            Nincr = np.repeat(1/Nstep, Nstep)
+            x = np.cumsum(Nincr)
+        # Interpolate multiplying number of points
+        elif Ninterp > 1:
+            x = np.linspace(0, az.size-1, (az.size-1)*Ninterp)
+        # Interpolate
+        if (DAinterp is not None) or (Ninterp > 1):
+            xp = np.arange(az.size)
+            az = np.interp(x, xp, az)
+            el = np.interp(x, xp, el)
+            if param is not None:
+                param = np.interp(x, xp, param)
+            if depol:
+                pol = np.interp(x, xp, pol)
+
+    # Calculate cartesian points.
+    x, y, z = azel_2_xyz(az, el, pol=pol)
+
+    # Return
+    if in_degrees:
+        ret = [x, y, z, az/degrees, el/degrees, param]
+    else:
+        ret = [x, y, z, az, el, param]
+    return ret
+
+
+def azel_2_xyz(az, el, pol=1):
+    """Transforms azimuth and ellipticity angle to x, y and z coordinates of the Poincare sphere.
+
+    Args:
+        az (np.ndarray): Azimuth.
+        el (np.ndarray): Ellipticity angle.
+        pol (np.ndarray): Polarization degree.
+
+    Returns:
+        x, y, z (np.ndarray): Calculated coordinates.
+    """
+    phi = 2*az
+    chi = -2*el + 90*degrees
+    x = pol * np.sin(chi) * np.cos(phi)
+    y = pol * np.sin(chi) * np.sin(phi)
+    z = pol * np.cos(chi)
+    return x, y, z
+
+
 # recreate the fitted curve using the optimized parameters
